from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self, Literal
    from .color import HexColorRange

import os
from typing import overload

from .dotset import Dotset


class DotsetLib(object):

    def __new__(cls):
        return cls.__new(())

    @classmethod
    def __new(cls, libdata: tuple[Dotset, ...]) -> Self:
        self = super().__new__(cls)
        self.__init(libdata)
        return self

    def __init(self, libdata: tuple[Dotset, ...]):
        self.__libdata = libdata
        self.__groups = {}
        isfontlib = isinstance(self, FontLib)
        for dotset in libdata:
            if isfontlib and len(dotset.name) != 1:
                raise KeyError(
                    f'Only dotset with 1-length-name can be stored in FontLib, "{dotset.name}" got')
            self.__groups[dotset.name] = self.__groups.get(dotset.name, ()) + (dotset,)
        if isfontlib: self.__charset = ''.join(self.__groups)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}({self.ngroup}/{self.ndotset}) at {hex(id(self))}>'

    def __len__(self) -> int:
        return self.__libdata.__len__()

    def __getitem__(self, item) -> Dotset:
        return self.__libdata[item]


    @classmethod
    def fromfile(cls, filepath: str) -> Self:
        with open(filepath, 'r') as f:
            datas = f.read().strip().split('\n')
        return cls.__new(tuple(filter(lambda item:item, (Dotset.fromdata(data) for data in datas))))

    def tofile(self, filepath: str) -> str:
        savedata = '\n'.join([dotset.todata() for dotset in self.__libdata])
        with open(filepath, 'w') as f:
            f.write(savedata)
        return os.path.abspath(filepath)


    @property
    def ndotset(self):
        return self.__libdata.__len__()

    @property
    def ngroup(self):
        return self.__groups.__len__()

    def group(self, name: str) -> tuple[Dotset, ...]:
        return self.__groups.get(name, ())

    def groups(self) -> dict[str, tuple[Dotset, ...]]:
        return self.__groups.copy()


    @overload
    def scale(self, scale: int | float) -> Self: ...

    @overload
    def scale(self, scaleX: int | float, scaleY: int | float) -> Self: ...

    def scale(self, scaleX: int | float, scaleY: int | float = None) -> Self:
        return self.__class__.__new(tuple(dotset.scale(scaleX, scaleY) for dotset in self.__libdata))


    def list(self) -> None:
        maxLabelLen = (self.ndotset-1).__str__().__len__()
        seperator = "=" * 60
        print(f'\n{seperator}')
        for i, dotset in enumerate(self.__libdata):
            print(f'{i:{maxLabelLen}}: {dotset}')
        print(f'{seperator}\n')

    def sort(self) -> None:
        self.__init(tuple(sorted(self.__libdata, key=lambda dotset: dotset.name)))

    def add(self, dotset: Dotset, toindex: int = None) -> None:
        if toindex is None:
            toindex = self.ndotset
        self.__init((*self.__libdata[:toindex], dotset, *self.__libdata[toindex:]))

    def delete(self, index: int) -> None:
        self.__init((*self.__libdata[:index], *self.__libdata[index+1 or self.ndotset:]))

    def __replace(self, index: int, dotset: Dotset) -> None:
        self.__init((*self.__libdata[:index], dotset, *self.__libdata[index+1 or self.ndotset:]))

    def rename(self, index: int, name: str) -> None:
        self.__replace(index, self[index].asname(name))

    def recolor(self, index: int, rgbrange: Literal[0,1] | HexColorRange) -> None:
        self.__replace(index, self[index].asmatchcolor(rgbrange))


class FontLib(DotsetLib):
    @property
    def charset(self): return self._DotsetLib__charset



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self, Iterable
    from .dotset import Dotset

import cv2 as cv
import numpy as np

from .target import Target, TargetList


class DsetSimMat(object):

    def __new__(cls):
        raise TypeError(
            'cannot create DsetSimMat instances')

    @classmethod
    def _new_(cls, dsetlist: Iterable[Dotset], simmat: np.ndarray, indexmat: np.ndarray = None) -> Self:
        self = super().__new__(cls)
        self._init_(dsetlist, simmat, indexmat)
        return self

    def _init_(self, dsetlist: Iterable[Dotset], simmat: np.ndarray, indexmat: np.ndarray = None):
        simmat[simmat == np.inf] = -np.inf
        self._dsetlist = tuple(dsetlist)
        self._simmat = simmat
        self._indexmat = np.zeros(simmat.shape, dtype=int) if indexmat is None else indexmat

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.size[0]}x{self.size[1]} at {hex(id(self))}>'


    @classmethod
    def merge(cls, dsetsimmat: Self, *dsetsimmats: Self) -> Self:
        if not dsetsimmats:
            return dsetsimmat
        in_dsetsimmats = (dsetsimmat, *dsetsimmats)
        out_dsetlist = ()
        out_matshape = max(mat.size[1] for mat in in_dsetsimmats), max(mat.size[0] for mat in in_dsetsimmats)
        out_simmat, out_indexmat = np.zeros(out_matshape), np.zeros(out_matshape, dtype=int)
        for dsetsimmat in in_dsetsimmats:
            dsetlist, simmat, indexmat = dsetsimmat._dsetlist, dsetsimmat._simmat, dsetsimmat._indexmat
            viewslice = slice(None, simmat.shape[0]), slice(None, simmat.shape[1])
            out_simmat_view, out_indexmat_view = out_simmat[viewslice], out_indexmat[viewslice]
            _filter = out_simmat_view < simmat
            out_simmat_view[_filter] = simmat[_filter]
            out_indexmat_view[_filter] = indexmat[_filter] + len(out_dsetlist)
            out_dsetlist += dsetlist
        return cls._new_(out_dsetlist, out_simmat, out_indexmat)


    @property
    def size(self) -> tuple[int, int]:
        return self._simmat.shape[1::-1]

    def get(self, x, y) -> tuple[Dotset, float]:
        return self.getdotset(x,y), self.getsim(x,y)

    def getsim(self, x, y) -> float:
        return self._simmat[y,x]

    def getdotset(self, x, y) -> Dotset:
        return self._dsetlist[self._indexmat[y,x]]


    def print_sims(self):
        for line in self._simmat:
            print(line.tolist())

    def print_dotsets(self):
        for line in self._indexmat:
            print([self._dsetlist[i] for i in line])


    def __filter(self, similarity, ignore_overlaps=False, direction=0) -> tuple[np.ndarray, np.ndarray]:
        _filter = self._simmat >= similarity
        if direction==0: ylist, xlist = np.where(_filter)
        if direction==1: xlist, ylist = np.where(_filter.T)
        if ignore_overlaps:
            for x0, y0 in zip(xlist, ylist):
                if not _filter[y0, x0]: continue
                dsetw, dseth = self.getdotset(x0, y0).size
                x1, y1 = x0 + dsetw, y0 + dseth
                localmaxX, localmaxY = cv.minMaxLoc(self._simmat[y0:y1, x0:x1])[-1]
                _filter[y0:y1, x0:x1] = False
                _filter[y0+localmaxY, x0+localmaxX] = True
            if direction == 0: ylist, xlist = np.where(_filter)
            if direction == 1: xlist, ylist = np.where(_filter.T)
        return xlist, ylist

    def filter(self, similarity, ignore_overlaps=False, direction=0) -> TargetList:
        xlist, ylist = self.__filter(similarity, ignore_overlaps, direction)
        return TargetList._new_((xlist, ylist, lambda x,y: Target._DotsetTarget((x,y), *self.get(x,y))))

    def separate(self, similarity, ignore_overlaps=False, direction=0) -> list[TargetList]:
        xlist, ylist = self.__filter(similarity, ignore_overlaps, direction)
        groups, groupBorderEnd = [], -1
        for pos in zip(xlist, ylist):
            dotset, sim = self.getdotset(*pos), self.getsim(*pos)
            if pos[not direction] > groupBorderEnd:
                groups.append([])
            groups[-1].append(Target._DotsetTarget(pos, dotset, sim))
            groupBorderEnd = max(groupBorderEnd, pos[not direction] + dotset.size[not direction])
        for i in range(len(groups)):
            groups[i].sort(key=lambda target: target.start[direction])
            groups[i] = TargetList._new_((groups[i],))
        return groups



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal, Iterator

import re
import numpy as np

from typing import Union

Uint8, Uhex8 = int, str
TupColor = Union[ tuple[Uint8], tuple[Uint8, Uint8, Uint8], tuple[Uint8, Uint8, Uint8, Uint8]]
HexColor = str          # see TypeError in matchHexColor()
HexColorRange = str     # see TypeError in matchHexColorRange()


def matchHexColor(color: HexColor):
    uhex8 = r'([0-9a-fA-F]{2})'
    match = re.search(f'^{uhex8}(?:{uhex8}{uhex8}{uhex8}?)?$', color)
    if not match: raise TypeError(
        f"HexColor shuold be a string like '0f', '0f0f0f' or '0f0f0fff'. got '{color}'")
    return match.groups()   # (r, g, b, a), None will replace blank

def matchHexColorRange(colorrange: HexColorRange):
    uhex8 = r'([0-9a-fA-F]{2})'
    match = re.search(f'^(-)?{uhex8}{uhex8}{uhex8}(?:(-{{1,2}}){uhex8}{uhex8}{uhex8})?$', colorrange)
    if not match: raise TypeError(
        "HexColorRange should be a string like '-?0e0e0e|-?0e0e0e(mid)-010101(diff)|-0d0d0d(min)--0f0f0f(max)', ")
    return match.groups()   # (sign, r1, g1, b1, sep, r2, g2, b2), None will replace blank

def hextup2inttup(uhex8tuple: tuple[Uhex8, ...]) -> tuple[Uint8, ...]:
    return tuple(int(uhex8, 16) for uhex8 in uhex8tuple if uhex8)


def rgb2gray(r: Uint8, g: Uint8, b: Uint8) -> Uint8:
    return int(0.114 * b + 0.587 * g + 0.299 * r + 0.5)

def translateRgb(rgbcolor: HexColor | TupColor, dstnchannel: Literal[1,3,4] = None) -> np.ndarray:
    if isinstance(rgbcolor, HexColor):
        rgbcolor = hextup2inttup(matchHexColor(rgbcolor))
    if dstnchannel and dstnchannel != len(rgbcolor):
        if dstnchannel == 1:
            rgbcolor = (rgb2gray(*rgbcolor[:3]),)
        else: rgbcolor = (*rgbcolor, *rgbcolor[:1]*(len(rgbcolor)==1)*2, 255)[:dstnchannel]
    bgrcolor = (*rgbcolor[2::-1], *rgbcolor[3:4])
    return np.array(bgrcolor, dtype=np.uint8)

def translateRgbrange(rgbrange: HexColorRange, graymode=False) -> Iterator[tuple[bool, TupColor, TupColor]]:
    mid_diff, min_max = 1, 2
    for rgbrange in rgbrange.split('|'):
        sign, r1, g1, b1, sep, r2, g2, b2 = matchHexColorRange(rgbrange)
        include, rangemode = sign != '-', len(sep or '--')
        lbgr, rbgr = hextup2inttup((b1, g1, r1)), hextup2inttup((b2, g2, r2))
        if rangemode == min_max:
            minbgr, maxbgr = lbgr, rbgr or lbgr
        if rangemode == mid_diff:
            midbgr, difbgr = lbgr, rbgr or (0, 0, 0)
            minbgr = tuple(max(midbgr[i] - difbgr[i], 0x00) for i in range(3))
            maxbgr = tuple(min(midbgr[i] + difbgr[i], 0xff) for i in range(3))
        if graymode:
            minbgr, maxbgr = (max(minbgr),), (min(maxbgr),)
        yield include, minbgr, maxbgr


def isColorsInRgbrange(immat: np.ndarray, rgbrange: HexColorRange) -> np.ndarray:
    isInRanges, graymode = None, immat.ndim == 2
    if not graymode: immat = immat[:,:,:3]
    for include, minbgr, maxbgr in translateRgbrange(rgbrange, graymode):
        isInRange = (immat >= minbgr) * (immat <= maxbgr)
        if not graymode:
            isInRange = isInRange.all(2)
        if not include:
            isInRange = ~isInRange
        isInRanges = isInRange if isInRanges is None else np.bitwise_or(isInRanges, isInRange)
    return isInRanges



from aclib.cv import DotsetLib, Image

dl = DotsetLib.fromfile(r"E:\#project\python\PersonalLib\.test\ddd")

for d in dl:
    d.print()
r = Image.fromfile('ym.png').finddotset(dl, 'ICON_频道筛选', similarity=0.65)
print(r.offset(0, -30))



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self, Callable
    from .dotset import Dotset

from aclib.builtins.__API__._pixel_target import BaseTarget


class Target(BaseTarget):

    @classmethod
    def _DotsetTarget(cls, pos: tuple[int,int], dotset: Dotset, similarity) -> Self:
        return cls(dotset.name, pos, (pos[0]+dotset.size[0], pos[1]+dotset.size[1]), similarity)


class TargetList(object):

    def __new__(cls):
        return cls._new_(([],))

    @classmethod
    # TargetList 存储原始数据, 只有在需要的时候才读取或遍历 Target, 在项目较多时可以显著提升速度和节省资源
    def _new_(cls, *logicgroups: tuple[list[Target]] | tuple[ list[int], list[int], Callable[[int,int], Target] ]) -> Self:

        """ logicgroup: ([Target, ...], ) | ( [x1, x2, ...], [y1, y2, ...], (x,y)=>Target ) """

        self = super().__new__(cls)
        self._init_(*logicgroups)
        return self

    def _init_(self, *logicgroups):
        self._logicgroups = [*logicgroups]

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} x{self.__len__()} {list(self)}>'

    def __bool__(self) -> bool:
        return bool(self.__len__())

    def __len__(self) -> int:
        return sum(len(logicgroup[0]) for logicgroup in self._logicgroups)

    def __getitem__(self, index) -> Target:
        if index < 0:
            index = self.__len__() + index
        if index >= 0:
            for logicgroup in self._logicgroups:
                targetnum = len(logicgroup[0])
                if index < targetnum:
                    if len(logicgroup) == 1:
                        return logicgroup[0][index]
                    return logicgroup[2](logicgroup[0][index], logicgroup[1][index])
                index -= targetnum
        raise IndexError(
            'targetlist index out of range')

    def __add__(self, other) -> Self:
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for +: '{self.__class__.__name__}' and '{type(other).__name__}'")
        return self.__class__._new_(*self._logicgroups, *other._logicgroups)


    def append(self, target: Target):
        if len(self._logicgroups[-1]) == 3:
            self._logicgroups.append(([],))
        self._logicgroups[-1][0].append(target)

    def join(self, start: int = 0, end: int = None) -> Target:
        selflen = len(self)
        if end is None:
            end = selflen
        if start < 0:
            start = max(0, selflen + start)
        if end < 0:
            end = max(0, selflen + end)
        start, end = min(start, selflen), min(end, selflen)
        if start < end:
            targets = [self[i] for i in range(start, end)]
            name = ''.join(map(lambda t: t.name, targets))
            avgsim = sum(map(lambda t: t.similarity, targets)) / len(targets)
            return Target(name, targets[0].start, targets[-1].end, avgsim)
        else: return Target.none



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal, Iterable, Self, Sequence
    from .dotsetlib import DotsetLib, FontLib
    from .color import TupColor, HexColor, HexColorRange

import re, os, tempfile
import cv2 as cv, numpy as np

from typing import overload

from .dsetsimmat import DsetSimMat
from .target import Target, TargetList
from .color import translateRgb, isColorsInRgbrange, rgb2gray

class Image(object):

    def __new__(cls):
        raise TypeError(
            "to create a 'Image' instance, please use Image.fromxxx method")

    @classmethod
    def __new(cls, immat: np.ndarray, imsize: tuple[int,int] = None) -> Self | None:
        if not immat.size: return None
        self = super().__new__(cls)
        self.__init(immat, imsize)
        return self

    def __init(self, immat: np.ndarray, imsize: tuple[int,int] = None):
        if imsize:
            npixel = imsize[0] * imsize[1]
            nchannel = npixel and immat.size // npixel
            ndim = 2 + nchannel // 3
            immat = immat.reshape((*imsize[1::-1], nchannel)[:ndim])
        if immat.ndim not in [2, 3]:
            raise TypeError('demensions of the array which describes the image must be 2 or 3')
        if immat.ndim == 3 and immat.shape[2] not in [1,3,4]:
            raise TypeError('image channels must be 1, 3, or 4')
        if immat.ndim == 3 and immat.shape[2] == 1:
            immat = immat.reshape((immat.shape[:2]))
        self._immat = immat

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.width}x{self.height}#{self.nchannel} at {hex(id(self))}>'

    def __mul__(self, other) -> Image:
        return self.__class__.__new(self._immat * other)


    @classmethod
    @overload
    def fromarray(cls, immat) -> Self | None: ...

    @classmethod
    @overload
    def fromarray(cls, array, imsize: tuple[int,int]) -> Self | None: ...

    @classmethod
    def fromarray(cls, array, imsize: tuple[int,int] = None) -> Self | None:
        return cls.__new(np.array(array, dtype=np.uint8), imsize)

    @classmethod
    def frombuffer(cls, buffer: bytes | bytearray, imsize: tuple[int,int]) -> Self | None:
        return cls.__new(np.frombuffer(buffer, dtype=np.uint8), imsize)

    @classmethod
    def fromiter(cls, data: Iterable, imsize: tuple[int,int]) -> Self | None:
        return cls.__new(np.fromiter(data, dtype=np.uint8), imsize)

    @classmethod
    def fromcolor(cls, rgbcolor: HexColor | TupColor, imsize: tuple[int,int]) -> Self | None:
        color = translateRgb(rgbcolor)
        ndim = 2 + len(color) // 3
        return cls.__new(np.tile(color, (imsize[1], imsize[0], 1)[:ndim]))

    @classmethod
    def fromfile(cls, imgpath: str) -> Self | None:
        imbuffer, immat = np.fromfile(imgpath, dtype=np.uint8), None
        if imbuffer.size:
            immat = cv.imdecode(imbuffer, cv.IMREAD_UNCHANGED)
        if immat is None:
            raise TypeError('invallid image file')
        return cls.__new(immat)

    def copy(self) -> Self:
        return self.__class__.__new(self._immat.copy())

    def tofile(self, savepath: str) -> str:
        cv.imencode(f'.{savepath.rsplit(".",1)[-1]}', self._immat)[1].tofile(savepath)
        return os.path.abspath(savepath)

    def toarray(self) -> np.ndarray:
        return self._immat.copy()

    def tolist(self) -> list:
        return self._immat.tolist()


    @property
    def width(self) -> int:
        return self._immat.shape[1]

    @property
    def height(self) -> int:
        return self._immat.shape[0]

    @property
    def nchannel(self):
        if self._immat.ndim == 2: return 1
        return self._immat.shape[2]

    @property
    def npixel(self):
        return self._immat.shape[1] * self._immat.shape[0]

    @property
    def size(self) -> tuple[int,int]:
        return self._immat.shape[1::-1]

    @property
    def shape(self) -> tuple[int,int] | tuple[int,int,Literal[3,4]]:
        return self._immat.shape


    def show(self, usebrowser=True):
        if usebrowser:
            htmp, tmpfname = tempfile.mkstemp(suffix='.png')
            os.close(htmp)
            self.tofile(tmpfname)
            os.system(tmpfname)
            os.remove(tmpfname)
        else:
            cv.imshow(str(self), self._immat)
            cv.waitKey()
            cv.destroyAllWindows()

    def printsrc(self):
        displaymat = self._immat.astype(str)
        print('[')
        if displaymat.ndim == 2:
            for line in displaymat:
                print(f' [{", ".join([v.rjust(3) for v in line])}]')
        if displaymat.ndim == 3:
            for line in displaymat:
                print(' [', end='')
                for col in line:
                    print(f'[{",".join([v.rjust(3) for v in col])}]', end=', ')
                print('\b\b]')
        print(']')

    def printbinimg(self, mask: Image = None):
        assert self.nchannel==1 and ((self._immat==0)+(self._immat==255)).all(), 'Image must be binarized'
        displaymat = self._immat.astype(str)
        displaymat[displaymat[:,:]=='0'] = '··'
        displaymat[displaymat[:,:]=='255'] = '##'
        if mask: displaymat[mask._immat[:,:]==0] = '  '
        sep = '==' * (self.width+1)
        print(sep)
        for line in displaymat: print('|' + "".join(["".join(color) for color in line]) + '|')
        print(f'{sep}\n')


    def swapchannel(self) -> Self:
        nchannel = self.nchannel
        if nchannel == 1:
            return self.copy()
        return self.__class__.__new(self._immat[:,:,(2,1,0,*[3]*(nchannel==4))])

    def tonchannel(self, nchannel: Literal[1,3,4]) -> Self:
        modes = {1: 'GRAY', 3: 'BGR', 4: 'BGRA'}
        frommode = modes.get(self.nchannel)
        tomode = modes.get(nchannel) or f'channel{nchannel}'
        if frommode != tomode:
            transmode = getattr(cv, f'COLOR_{frommode}2{tomode}')
            return self.__class__.__new(cv.cvtColor(self._immat, transmode))
        return self.copy()

    def binarize(self, rgbrange: Literal[0,1] | HexColorRange) -> Self:
        if rgbrange == 0 or rgbrange == 1:
            grayscale = self._immat if self._immat.ndim==2 else cv.cvtColor(self._immat[:,:,:3], cv.COLOR_BGR2GRAY)
            binmat = cv.threshold(grayscale, 0, 255, ~rgbrange+2 | cv.THRESH_OTSU)[1]
        if isinstance(rgbrange, str):
            binmat = isColorsInRgbrange(self._immat, rgbrange).astype(np.uint8) * 255
        return self.__class__.__new(binmat)


    @overload
    def scale(self, scale: int | float) -> Self: ...

    @overload
    def scale(self, scaleX: int | float, scaleY: int | float) -> Self: ...

    def scale(self, scaleX: int | float, scaleY: int | float = None) -> Self:
        if scaleY is None: scaleY = scaleX
        assert scaleX > 0 and scaleY > 0, 'scale argument must be a positive number'
        if not self: return self.copy()
        return self.__class__.__new(cv.resize(self._immat, None, None, scaleX, scaleY, cv.INTER_AREA))


    def __bitoperation(self, method, *imgs):
        dstmat = self._immat
        for im in imgs: dstmat = method(dstmat, im._immat)
        return self.__class__.__new(dstmat)

    def bitwise_or(self, img: Self, *imgs: Self) -> Self:
        return self.__bitoperation(cv.bitwise_or, img, *imgs)

    def bitwise_xor(self, img: Self, *imgs: Self) -> Self:
        return self.__bitoperation(cv.bitwise_xor, img, *imgs)


    def minrgb(self) -> np.ndarray:
        if self.nchannel == 1:
            return np.repeat(self._immat.min(), 3)
        return self._immat[::3].min(0).min(0)[2::-1]

    def maxrgb(self) -> np.ndarray:
        if self.nchannel == 1:
            return np.repeat(self._immat.max(), 3)
        return self._immat[::3].max(0).max(0)[2::-1]

    def mainrgbrange(self, uint8offset: int = 0) -> HexColorRange:
        firstrgb = np.repeat(self._immat[0,0], 3) if self.nchannel==1 else self._immat[0,0][2::-1]
        diffrgb = np.repeat(uint8offset, 3)
        minrgb = self.minrgb()
        maxrgb = self.maxrgb()
        if rgb2gray(*firstrgb) <= rgb2gray(*minrgb)/2 + rgb2gray(*maxrgb)/2:
            minrgb = [max(minrgb[i], maxrgb[i] - diffrgb[i]) for i in range(3)]
        else: maxrgb = [min(maxrgb[i], minrgb[i] + diffrgb[i]) for i in range(3)]
        hexrgbs = ''.join(hex(v)[2:].zfill(2) for v in (*minrgb, *maxrgb))
        return f'{hexrgbs[:6]}--{hexrgbs[-6:]}'

    def border(self, objtype: Literal[0,1,2] = 0, objmargin: int = 0, bodrrgbrange: HexColorRange = '000000') \
            -> tuple[tuple[int,int], tuple[int,int]]:
        """
        提取图像中对象的边界
        :param objtype: 0:最小矩阵 1:全角矩阵 2:半角矩阵
        :param objmargin: 在对象四周保留的边距
        :param bodrrgbrange: 在图像四周与bodrcolor颜色相同的区域将被视为对象边框，以此判定对象的最小边界
        """
        srcH, srcW = self._immat.shape[:2]
        for col in range(srcW):
            if not isColorsInRgbrange(self._immat[:,col:col+1], bodrrgbrange).all(): break
        srcObjL = col
        for col in range(srcW - 1, srcObjL - 1, -1):
            if not isColorsInRgbrange(self._immat[:,col:col+1], bodrrgbrange).all(): break
        srcObjR = col + 1
        for line in range(srcH):
            if not isColorsInRgbrange(self._immat[line:line+1,:], bodrrgbrange).all(): break
        srcObjT = line
        for line in range(srcH - 1, srcObjT - 1, -1):
            if not isColorsInRgbrange(self._immat[line:line+1,:], bodrrgbrange).all(): break
        srcObjB = line + 1
        if objtype == 0:
            dstObjL, dstObjR, dstObjT, dstObjB = srcObjL, srcObjR, srcObjT, srcObjB
        if objtype in [1, 2]:
            srcObjW, srcObjH = srcObjR - srcObjL, srcObjB - srcObjT
            srcObjHWRatio = srcObjH / srcObjW
            dstObjHWRatio = objtype
            dstObjW = int(srcObjH / dstObjHWRatio) if srcObjHWRatio >= dstObjHWRatio else srcObjW
            dstObjH = srcObjH                      if srcObjHWRatio >= dstObjHWRatio else srcObjW * dstObjHWRatio
            paddingX, paddingY = int(dstObjW / 2 - srcObjW / 2), int(dstObjH / 2 - srcObjH / 2)
            if srcObjL < paddingX:
                dstObjL, dstObjR = srcObjL, min(srcObjL+dstObjW, srcW)
            else:
                dstObjR = min(srcObjR + paddingX, srcW)
                dstObjL = max(dstObjR-dstObjW, 0)
            if srcObjT < paddingY:
                dstObjT, dstObjB = srcObjT, min(srcObjT+dstObjH, srcH)
            else:
                dstObjB = min(srcObjB+paddingY, srcH)
                dstObjT = max(dstObjB-dstObjH, 0)
        if objmargin >= 0:
            dstL = max(dstObjL - objmargin, 0)
            dstR = min(dstObjR + objmargin, srcW)
            dstT = max(dstObjT - objmargin, 0)
            dstB = min(dstObjB + objmargin, srcH)
        return (dstL, dstT), (dstR, dstB)


    @classmethod
    def __concat(cls, imgs: Sequence[Self], fillrgb: HexColor | TupColor, axis: Literal[0,1]) -> Self:
        maxnchannel = max(im.nchannel for im in imgs)
        maxsize_onvaxis = max(im.shape[not axis] for im in imgs)
        imgs = map(lambda im:im if im.nchannel==maxnchannel else im.tonchannel(maxnchannel), imgs)
        fillcolor, fillndim = translateRgb(fillrgb, maxnchannel), 2 + maxnchannel // 3
        def fillmatof(im):
            imh, imw = im.shape[:2]
            reps = [(imh, maxsize_onvaxis - imw, 1), (maxsize_onvaxis-imh, imw, 1)][axis][:fillndim]
            return np.tile(fillcolor, reps)
        imgs = [np.concatenate([im._immat, fillmatof(im)], ~axis+2) for im in imgs]
        return cls.__new(np.concatenate(imgs, axis))

    @classmethod
    def concatH(cls, himgs: Sequence[Self], fillrgb: HexColor | TupColor = (0,)) -> Self:
        return cls.__concat(himgs, fillrgb, 1)

    @classmethod
    def concatV(cls, vimgs: Sequence[Self], fillrgb: HexColor | TupColor = (0,)) -> Self:
        return cls.__concat(vimgs, fillrgb, 0)

    @classmethod
    def concatHV(cls, himgslist: Sequence[Sequence[Self]], fillrgb: HexColor | TupColor = (0,)) -> Self:
        return cls.__concat([cls.__concat(himgs, fillrgb, 1) for himgs in himgslist], fillrgb, 0)

    @classmethod
    def concatVH(cls, vimgslist: Sequence[Sequence[Self]], fillrgb: HexColor | TupColor = (0,)) -> Self:
        return cls.__concat([cls.__concat(vimgs, fillrgb, 0) for vimgs in vimgslist], fillrgb, 1)

    def crop(self, start: tuple[int,int], end: tuple[int,int]) -> Self:
        l = max(start[0], 0)
        t = max(start[1], 0)
        r = min(end[0], self.width)
        b = min(end[1], self.height)
        return self.__class__.__new(self._immat[t:b, l:r].copy())

    def removeborder(self, bodrrgbrange: HexColorRange = '000000--303030') -> Self:
        return self.crop(*self.border(0, 0, bodrrgbrange))

    @overload
    def addborder(self, bodrwidth: int, rgbbodrcolor: HexColor | TupColor = (0,)) -> Self: ...

    @overload
    def addborder(self, bodrX: int, bodrY: int, rgbbodrcolor: HexColor | TupColor = (0,)) -> Self: ...

    @overload
    def addborder(self, bodrL: int, bodrY: int, bodrR: int, rgbbodrcolor: HexColor | TupColor = (0,)) -> Self: ...

    @overload
    def addborder(self, bodrL: int, bodrT: int, bodrR: int, bodrB: int, rgbbodrcolor: HexColor | TupColor = (0,)) -> Self: ...

    def addborder(self, *args) -> Self:
        if args and not hasattr(args[-1], '__index__'):
            args, rgbbodrcolor = args[:-1], args[-1]
        else: rgbbodrcolor = (0,)
        if not 1 <= len(args) <= 4:
            raise TypeError(
                'addborder() takes only 1-4 bodrwidth and 1 bodrcolor')
        lw, tw, rw, bw = [*args*(4//len(args)), *args[1:2]][:4]
        bodrcolor = translateRgb(rgbbodrcolor, self.nchannel)
        selfw, selfh = self.size
        selfndim = self._immat.ndim
        bodrmat = lambda h,w: np.tile(bodrcolor, (h,w,1)[:selfndim])
        lmat, rmat = bodrmat(selfh, lw), bodrmat(selfh, rw)
        tmat, bmat = bodrmat(tw, selfw+lw+rw), bodrmat(bw, selfw+lw+rw)
        return self.__class__.__new(np.concatenate([tmat, np.concatenate([lmat, self._immat, rmat], 1), bmat], 0))


    def findcolor(self, rgbrange: HexColorRange) -> Target:
        for rgbrange in rgbrange.split('|'):
            ylist, xlist = np.where(isColorsInRgbrange(self._immat, rgbrange))
            if len(xlist):
                pos = xlist[0], ylist[0]
                return Target(rgbrange, pos, pos, 1.0)
        return Target.none

    def findcolors(self, rgbrange: HexColorRange) -> TargetList:
        targets = TargetList()
        for rgbrange in rgbrange.split('|'):
            ylist, xlist = np.where(isColorsInRgbrange(self._immat, rgbrange))
            targets += TargetList._new_((xlist, ylist, lambda x,y: Target(rgbrange, (x,y), (x,y), 1.0)))
        return targets


    def __matchdotset(self, dotset, matchcolor, dsetscale, cache) -> DsetSimMat:
        # 在 ocr 和 finddotset 中调用，每次调用除了 dotset 外，其它参数是相同的
        # cache: {color: binimage, dotset: matchresult}
        # 点阵缩放后的宽和高不应大于图片，即使其宽高同时大于图片时不会抛错，但那样的匹配结果应该是毫无意义的
        if dotset not in cache:
            if matchcolor is None:
                matchcolor = dotset.matchcolor
            if matchcolor not in cache:
                cache[matchcolor] = self.binarize(matchcolor)
            scaleddset = dotset.scale(dsetscale)
            simmat = cv.matchTemplate(
                cache[matchcolor]._immat,
                scaleddset.tmpl._immat,
                cv.TM_CCOEFF_NORMED, None,
                scaleddset.mask and scaleddset.mask._immat
            )[:self.height, :self.width]
            cache[dotset] = DsetSimMat._new_([scaleddset], simmat)
        return cache[dotset]

    def finddotset(self,
        dotsetlib: DotsetLib,
        *dotset: str,
        matchcolor: Literal[0,1] | HexColorRange = None,
        similarity = 0.95,
        dsetscale = 1.0
    ) -> Target:
        cache = {}
        for name in dict.fromkeys(dotset):
            for dotset in dotsetlib.group(name):
                targets = self.__matchdotset(dotset, matchcolor, dsetscale, cache).filter(similarity)
                if targets: return targets[0]
        return Target.none

    def finddotsets(self,
        dotsetlib: DotsetLib,
        *dotset: str,
        matchcolor: Literal[0,1] | HexColorRange = None,
        similarity = 0.95,
        dsetscale = 1.0,
        ignore_overlaps = False
    ) -> TargetList:
        cache = {}
        targets = TargetList()
        for name in dict.fromkeys(dotset):
            dsetsimmats = []
            for dotset in dotsetlib.group(name):
                dsetsimmats.append(self.__matchdotset(dotset, matchcolor, dsetscale, cache))
            if dsetsimmats:
                targets += DsetSimMat.merge(*dsetsimmats).filter(similarity, ignore_overlaps)
        return targets

    def __ocr(self, fontlib, matchcolor, similarity, txtdir, txtwrap, charscale, charset, cache) -> list[TargetList]:
        dsetsimmats = []
        if charset != fontlib.charset:
            charset = dict.fromkeys(charset)
        for char in charset:
            for dotset in fontlib.group(char):
                dsetsimmats.append(self.__matchdotset(dotset, matchcolor, charscale, cache))
        if dsetsimmats:
            chargroups = DsetSimMat.merge(*dsetsimmats).separate(similarity, True, txtdir)
            if txtwrap: chargroups = [sum(chargroups, TargetList())]
            return chargroups
        return []

    def ocr(self,
        fontlib: FontLib,
        matchcolor: Literal[0,1] | HexColorRange = None,
        similarity = 0.95,
        txtdir: Literal[0,1] = 0,
        txtwrap = False,
        charscale = 1.0,
        charset: str | Literal[''] = None
    ) -> list[TargetList]:
        cache = {}
        return self.__ocr(fontlib, matchcolor, similarity, txtdir, txtwrap, charscale, charset or fontlib.charset, cache)

    def findtext(self,
        fontlib: FontLib,
        *text: str,
        matchcolor: Literal[0,1] | HexColorRange = None,
        similarity = 0.95,
        txtdir: Literal[0,1] = 0,
        txtwrap = False,
        charscale = 1.0,
        charset: str | Literal[''] = None
    ) -> Target:
        cache = {}
        ocr = lambda _charset: self.__ocr(fontlib, matchcolor, similarity, txtdir, txtwrap, charscale, _charset, cache)
        if charset != '': chargroups = ocr(charset or fontlib.charset)
        for text in text:
            if charset == '': chargroups = ocr(text)
            for charlist in chargroups:
                index = charlist.join().name.find(text)
                if index >= 0: return charlist.join(index, index + len(text))
        return Target.none

    def findtexts(self,
        fontlib: FontLib,
        *text: str,
        matchcolor: Literal[0,1] | HexColorRange = None,
        similarity = 0.95,
        txtdir: Literal[0,1] = 0,
        txtwrap = False,
        charscale = 1.0,
        charset: str | Literal[''] = None,
        ignore_overlaps = False
    ) -> TargetList:
        cache, found = {}, TargetList()
        ocr = lambda _charset: self.__ocr(fontlib, matchcolor, similarity, txtdir, txtwrap, charscale, _charset, cache)
        if charset != '': chargroups = ocr(charset or fontlib.charset)
        for text in text:
            if text == '': continue
            if charset == '': chargroups = ocr(text)
            txtlen, txtreg = len(text), re.escape(text)
            regex = txtreg if ignore_overlaps else f'(?=({txtreg}))'
            for charlist in chargroups:
                for match in re.finditer(regex, charlist.join().name):
                    index = match.start()
                    found.append(charlist.join(index, index + txtlen))
        return found



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal, Self
    from .color import HexColorRange

from typing import overload
from random import randint
from aclib.builtins import Str

from .image import Image

class Dotset(object):

    def __new__(cls):
        raise TypeError(
            f"cannot create {cls.__name__} instances")

    @classmethod
    def __new(cls, name: str, bintmpl: Image, binmask: Image | None, rgbmatchcolor: Literal[0,1] | HexColorRange) -> Self:
        self = super().__new__(cls)
        self.__init(name, bintmpl, binmask, rgbmatchcolor)
        return self

    def __init(self, name, bintmpl, binmask, rgbmatchcolor):
        self.__name = name
        self.__tmpl = bintmpl
        self.__mask = binmask
        self.__matchcolor = rgbmatchcolor

    def __repr__(self) -> str:
        npixelinfo = f'{self.tmpl.npixel}' + f'-{self.npixel}' * (self.mask is not None)
        return f'<Dotset:{self.name} {self.matchcolor}|{self.width}x{self.height}={npixelinfo}>'


    @classmethod
    def fromimage(cls,
        asname:              str,
        image:               Image,
        tmplExtractColor:    Literal[0,1] | HexColorRange,
        maskExtractColor:    Literal[0,1] | HexColorRange = None,
        matchcolor:          Literal[0,1] | HexColorRange = None,
        cropmode:            Literal[0,1,2] = 0,
        cropmargin:          int = 0,
    ) -> Self:
        uncroppedBintmpl = image.binarize(tmplExtractColor)
        if maskExtractColor is not None:
            uncroppedBinmask = image.binarize(maskExtractColor).bitwise_or(uncroppedBintmpl)
            cropStart, cropEnd = uncroppedBinmask.border(cropmode, cropmargin)
            bintmpl = uncroppedBintmpl.crop(cropStart, cropEnd)
            binmask = uncroppedBinmask.crop(cropStart, cropEnd)
        else: bintmpl, binmask = uncroppedBintmpl.crop(*uncroppedBintmpl.border(cropmode, cropmargin)), None
        if matchcolor is None:
            matchcolor = tmplExtractColor
        return cls.__new(asname, bintmpl, binmask, matchcolor)

    @classmethod
    def fromdata(cls, data: str) -> Self | None:
        try:
            def decode(zippedhexim: Str) -> Image | None:
                binim = zippedhexim.unzip('GHIJKLMNOP').toBin(16).data[-w*h:]
                return Image.fromiter(binim, (w,h)) * 255 if binim else None
            b64name, b64color, strw, strh, hextmpl, hexmask = tuple(Str(data).split('|'))   # to tuple to get type hint
            name, color = b64name.base64decode().tostr(), b64color.base64decode().tostr()
            if color in ['0', '1']: color = int(color)
            w, h = strw.toint(), strh.toint()
            tmpl, mask = decode(hextmpl), decode(hexmask)
            if mask: mask = mask.bitwise_or(tmpl)
            return cls.__new(name, tmpl, mask, color)
        except: return None

    def todata(self) -> str:
        def encode(mat) -> Str:
            return Str().join((mat/255).reshape(mat.size).astype(int).astype(str)).toHex(2).toStr().zip('GHIJKLMNOP')
        hextmpl = encode(self.tmpl.toarray())
        hexmask = encode(self.mask.bitwise_xor(self.tmpl).toarray()) if self.mask else chr(randint(ord('Q'), ord('Z')))
        b64name = Str(self.name).base64encode()
        b64color = Str(self.matchcolor).base64encode()
        return f'{b64name}|{b64color}|{self.width}|{self.height}|{hextmpl}|{hexmask}'


    @property
    def name(self) -> str:
        return self.__name

    @property
    def tmpl(self) -> Image:
        return self.__tmpl

    @property
    def mask(self) -> Image | None:
        return self.__mask

    @property
    def matchcolor(self) -> Literal[0,1] | HexColorRange:
        return self.__matchcolor

    @property
    def size(self) -> tuple[int,int]:
        return self.__tmpl.size

    @property
    def width(self) -> int:
        return self.__tmpl.width

    @property
    def height(self) -> int:
        return self.__tmpl.height

    @property
    def npixel(self) -> int:
        if self.mask:
            maskmat = self.mask._immat
            return maskmat[maskmat == 255].__len__()
        else: return self.tmpl.npixel


    def asname(self, name: str) -> Self:
        return self.__class__.__new(name, self.tmpl, self.mask, self.matchcolor)

    def asmatchcolor(self, rgbmatchcolor: Literal[0,1] | HexColorRange) -> Self:
        return self.__class__.__new(self.name, self.tmpl, self.mask, rgbmatchcolor)


    @overload
    def scale(self, scale: int | float) -> Self: ...

    @overload
    def scale(self, scaleX: int | float, scaleY: int | float) -> Self: ...

    def scale(self, scaleX: int | float, scaleY: int | float = None) -> Self:
        if scaleY is None: scaleY = scaleX
        if scaleX == scaleY == 1:
            return self
        scaledtmpl = self.tmpl.scale(scaleX, scaleY).binarize(1)
        scaledmask = self.mask and self.mask.bitwise_xor(self.tmpl).scale(scaleX, scaleY).binarize(1).bitwise_or(scaledtmpl)
        return self.__class__.__new(self.name, scaledtmpl, scaledmask, self.matchcolor)


    def print(self):
        print(self)
        self.__tmpl.printbinimg(self.mask)

    def show(self, usebrowser=True):
        displaymat = self.tmpl.toarray()
        if self.mask:
            displaymat[self.mask.toarray()[:,:] == 0] = 0xa0
        Image.fromarray(displaymat).show(usebrowser)
